@@ -17,11 +17,22 @@ Contributors:
 import asyncio
 import logging
 from types import TracebackType
-from typing import Any, Awaitable, Callable, Dict, Iterable, List, Optional, Type, Union
+from typing import (
+    Any,
+    Awaitable,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Optional,
+    Set,
+    Type,
+    Union,
+)
 
 from pypck import inputs, lcn_defs
 from pypck.lcn_addr import LcnAddr
-from pypck.module import GroupConnection, ModuleConnection
+from pypck.module import AbstractConnection, GroupConnection, ModuleConnection
 from pypck.pck_commands import PckGenerator
 
 _LOGGER = logging.getLogger(__name__)
@@ -251,9 +262,9 @@ class PchkConnectionManager(PchkConnection):
         # Tasks
         self.ping_task: Optional["asyncio.Task[None]"] = None
         self.read_data_loop_task: Optional["asyncio.Task[None]"] = None
-        self.request_serials_task: Optional[
+        self.request_serials_tasks: Set[
             "asyncio.Task[Dict[str, Union[int, lcn_defs.HardwareType]]]"
-        ] = None
+        ] = set()
 
         # Events, Futures, Locks for synchronization
         self.segment_scan_completed_event = asyncio.Event()
@@ -262,11 +273,12 @@ class PchkConnectionManager(PchkConnection):
         self.module_serial_number_received = asyncio.Lock()
         self.segment_coupler_response_received = asyncio.Lock()
 
-        # All modules/groups from or to a communication occurs are represented
-        # by a unique ModuleConnection or GroupConnection object.
-        # All ModuleConnection and GroupConnection objects are stored in this
-        # dictionary.
-        self.address_conns: Dict[LcnAddr, Union[ModuleConnection, GroupConnection]] = {}
+        # All modules from or to a communication occurs are represented by a
+        # unique ModuleConnection object.  All ModuleConnection objects are
+        # stored in this dictionary.  Communication to groups is handled by
+        # GroupConnection object that are created on the fly and not stored
+        # permanently.
+        self.address_conns: Dict[LcnAddr, ModuleConnection] = {}
         self.segment_coupler_ids: List[int] = []
 
     async def __aenter__(self) -> "PchkConnectionManager":
@@ -378,8 +390,7 @@ class PchkConnectionManager(PchkConnection):
         await super().async_close()
         if self.ping_task is not None:
             await cancel(self.ping_task)
-        if self.request_serials_task is not None:
-            await cancel(self.request_serials_task)
+        await asyncio.gather(*(cancel(t) for t in self.request_serials_tasks))
         await self.cancel_requests()
         _LOGGER.debug("Connection to %s closed.", self.connection_id)
 
@@ -411,7 +422,7 @@ class PchkConnectionManager(PchkConnection):
         :rtype:      :class:`~LcnAddr`
         """
         return LcnAddr(
-            self.local_seg_id if addr.seg_id == 0 else addr.seg_id,
+            self.local_seg_id if addr.seg_id in (0, 4) else addr.seg_id,
             addr.addr_id,
             addr.is_group,
         )
@@ -426,10 +437,67 @@ class PchkConnectionManager(PchkConnection):
         """
         return self.segment_scan_completed_event.is_set()
 
+    def get_module_conn(
+        self, addr: LcnAddr, request_serials: bool = True
+    ) -> ModuleConnection:
+        """Create and/or return the given LCN module.
+
+        The ModuleConnection object is used for further communication
+        with the module (e.g. sending commands).
+
+        :param    addr:    The module's address
+        :type     addr:    :class:`~LcnAddr`
+
+        :returns: The address connection object (never null)
+        :rtype: `~ModuleConnection`
+
+        :Example:
+
+        >>> address = LcnAddr(0, 7, False)
+        >>> module = pchk_connection.get_module_conn(address)
+        >>> module.toggle_output(0, 5)
+        """
+        assert not addr.is_group
+        if addr.seg_id == 0 and self.local_seg_id != -1:
+            addr = LcnAddr(self.local_seg_id, addr.addr_id, addr.is_group)
+        address_conn = self.address_conns.get(addr, None)
+        if address_conn is None:
+            address_conn = ModuleConnection(self, addr)
+            if request_serials:
+                request_task = asyncio.create_task(address_conn.request_serials())
+                self.request_serials_tasks.add(request_task)
+                request_task.add_done_callback(self.request_serials_tasks.remove)
+            self.address_conns[addr] = address_conn
+
+        return address_conn
+
+    def get_group_conn(self, addr: LcnAddr) -> GroupConnection:
+        """Create and return the GroupConnection for the given group.
+
+        The GroupConnection can be used for sending commands to all
+        modules that are static or dynamic members of the group.
+
+        :param    addr:    The group's address
+        :type     addr:    :class:`~LcnAddr`
+
+        :returns: The address connection object (never null)
+        :rtype: `~GroupConnection`
+
+        :Example:
+
+        >>> address = LcnAddr(0, 7, True)
+        >>> group = pchk_connection.get_group_conn(address)
+        >>> group.toggle_output(0, 5)
+        """
+        assert addr.is_group
+        if addr.seg_id == 0 and self.local_seg_id != -1:
+            addr = LcnAddr(self.local_seg_id, addr.addr_id, addr.is_group)
+        return GroupConnection(self, addr)
+
     def get_address_conn(
         self, addr: LcnAddr, request_serials: bool = True
-    ) -> Union[ModuleConnection, GroupConnection]:
-        """Create and/or return the given LCN module or group.
+    ) -> AbstractConnection:
+        """Create and/or return an AbstractConnection to the given module or group.
 
         The LCN module/group object is used for further communication
         with the module/group (e.g. sending commands).
@@ -438,30 +506,28 @@ class PchkConnectionManager(PchkConnection):
         :type     addr:    :class:`~LcnAddr`
 
         :returns: The address connection object (never null)
-        :rtype: `~ModuleConnection` or `~GroupConnection`
+        :rtype: `~AbstractConnection`
 
         :Example:
 
         >>> address = LcnAddr(0, 7, False)
-        >>> module = pchk_connection.get_address_conn(address)
-        >>> module.toggle_output(0, 5)
+        >>> target = pchk_connection.get_address_conn(address)
+        >>> target.toggle_output(0, 5)
         """
-        if addr.seg_id == 0 and self.local_seg_id != -1:
-            addr = LcnAddr(self.local_seg_id, addr.addr_id, addr.is_group)
-        address_conn = self.address_conns.get(addr, None)
-        if address_conn is None:
-            if addr.is_group:
-                address_conn = GroupConnection(self, addr)
-                self.address_conns[addr] = address_conn
-            else:
-                address_conn = ModuleConnection(self, addr)
-                if request_serials:
-                    self.request_serials_task = asyncio.create_task(
-                        address_conn.request_serials()
-                    )
-                self.address_conns[addr] = address_conn
+        if addr.is_group:
+            return self.get_group_conn(addr)
+        return self.get_module_conn(addr, request_serials)
 
-        return address_conn
+    def dump_modules(self) -> Dict[str, Dict[str, Dict[str, Any]]]:
+        """Dump all modules and information about them in a JSON serializable dict."""
+        dump: Dict[str, Dict[str, Dict[str, Any]]] = {}
+        for address_conn in self.address_conns.values():
+            seg = f"{address_conn.addr.seg_id:d}"
+            addr = f"{address_conn.addr.addr_id}"
+            if seg not in dump:
+                dump[seg] = {}
+            dump[seg][addr] = address_conn.dump_details()
+        return dump
 
     async def scan_modules(self, num_tries: int = 3, timeout_msec: int = 3000) -> None:
         """Scan for modules on the bus.
@@ -600,7 +666,8 @@ class PchkConnectionManager(PchkConnection):
         elif self.is_ready():
             assert isinstance(inp, inputs.ModInput)
             logical_source_addr = self.physical_to_logical(inp.physical_source_addr)
-            module_conn = self.get_address_conn(logical_source_addr)
+            assert not logical_source_addr.is_group
+            module_conn = self.get_module_conn(logical_source_addr)
             if isinstance(inp, inputs.ModSn):
                 # used to extend scan_modules() timeout
                 if self.module_serial_number_received.locked():
