@@ -1,9 +1,13 @@
-"""Copyright (c) 2006-2018 by the respective copyright holders.
+"""Copyright (c) 2006-2020 by the respective copyright holders.
 
-All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
-which accompanies this distribution, and is available at
-http://www.eclipse.org/legal/epl-v10.html
+See the NOTICE file(s) distributed with this work for additional
+information.
+
+This program and the accompanying materials are made available under the
+terms of the Eclipse Public License 2.0 which is available at
+http://www.eclipse.org/legal/epl-2.0
+
+SPDX-License-Identifier: EPL-2.0
 
 Contributors:
   Andre Lengwenus - port to Python and further improvements
@@ -13,13 +17,13 @@ Contributors:
 import math
 import re
 from enum import Enum, auto
-from itertools import product
+from typing import Any, Dict, Tuple, Union, no_type_check
 
 LCN_ENCODING = "utf-8"
 PATTERN_SPLIT_PORT_PIN = re.compile(r"(?P<port>[a-zA-Z]+)(?P<pin>\d+)")
 
 
-def split_port_pin(portpin):
+def split_port_pin(portpin: str) -> Tuple[str, int]:
     """Split the port and the pin from the given input string.
 
     :param    str    portpin:    Input string
@@ -111,18 +115,44 @@ class BinSensorPort(Enum):
     BINSENSOR8 = 7
 
 
-# Keys
-# pylint: disable=invalid-name
-Key = Enum(
-    "Key",
-    " ".join(
-        [
-            "{:s}{:d}".format(t[0], t[1])
-            for t in product(["A", "B", "C", "D"], range(1, 9))
-        ]
-    ),
-    module=__name__,
-)
+class Key(Enum):
+    """Keys of LCN module."""
+
+    A1 = 0
+    A2 = 1
+    A3 = 2
+    A4 = 3
+    A5 = 4
+    A6 = 5
+    A7 = 6
+    A8 = 7
+
+    B1 = 8
+    B2 = 9
+    B3 = 10
+    B4 = 11
+    B5 = 12
+    B6 = 13
+    B7 = 14
+    B8 = 15
+
+    C1 = 16
+    C2 = 17
+    C3 = 18
+    C4 = 19
+    C5 = 20
+    C6 = 21
+    C7 = 22
+    C8 = 23
+
+    D1 = 24
+    D2 = 25
+    D3 = 26
+    D4 = 27
+    D5 = 28
+    D6 = 29
+    D7 = 30
+    D8 = 31
 
 
 class OutputPortDimMode(Enum):
@@ -134,8 +164,8 @@ class OutputPortDimMode(Enum):
     Since LCN-PCHK doesn't know the current mode, it must explicitly be set.
     """
 
-    STEPS50 = auto()  # 0..50 dimming steps (all LCN module generations)
-    STEPS200 = auto()  # 0..200 dimming steps (since 170206)
+    STEPS50 = 0  # 0..50 dimming steps (all LCN module generations)
+    STEPS200 = 1  # 0..200 dimming steps (since 170206)
 
 
 class OutputPortStatusMode(Enum):
@@ -147,11 +177,11 @@ class OutputPortStatusMode(Enum):
     though.
     """
 
-    PERCENT = auto()  # Default (compatible with all versions of LCN-PCHK)
-    NATIVE = auto()  # 0..200 steps (since LCN-PCHK 2.3)
+    PERCENT = "P"  # Default (compatible with all versions of LCN-PCHK)
+    NATIVE = "N"  # 0..200 steps (since LCN-PCHK 2.3)
 
 
-def time_to_ramp_value(time_msec):
+def time_to_ramp_value(time_msec: int) -> int:
     """Convert the given time into an LCN ramp value.
 
     :param    int    time_msec:    The time in milliseconds.
@@ -180,13 +210,14 @@ def time_to_ramp_value(time_msec):
     elif time_msec < 6000:
         ret = 9
     else:
-        ret = (time_msec / 1000 - 6) / 2 + 10
-        if ret >= 250:
-            ret = 250
-    return int(ret)
+        ramp = (time_msec / 1000 - 6) / 2 + 10
+        if ramp >= 250:
+            ramp = 250
+        ret = int(ramp)
+    return ret
 
 
-def ramp_value_to_time(ramp_value):
+def ramp_value_to_time(ramp_value: int) -> int:
     """Convert the given LCN ramp value into a time.
 
     :param    int    ramp_value:    The LCN ramp value (0..250).
@@ -199,10 +230,54 @@ def ramp_value_to_time(ramp_value):
 
     if ramp_value < 10:
         times = [0, 250, 500, 660, 1000, 1400, 2000, 3000, 4000, 5000]
-        t = times[ramp_value]
+        ramp_time = times[ramp_value]
     else:
-        t = int(((ramp_value - 10) * 2 + 6) * 1000)
-    return t
+        ramp_time = int(((ramp_value - 10) * 2 + 6) * 1000)
+    return ramp_time
+
+
+def time_to_native_value(time_msec: int) -> int:
+    """Convert time to native LCN time value.
+
+    Scales the given time value in milliseconds to a byte value (0..255).
+    Used for RelayTimer.
+
+    :param    int    time_msec:    Duration of timer in milliseconds
+
+    :returns: The duration in native LCN units
+    :rtype: int
+    """
+    if not 0 <= time_msec <= 240960:
+        raise ValueError("Time has to be in range 0..240960ms.")
+    time_scaled = time_msec / (1000 * 0.03 * 32.0) + 1.0
+
+    pre_decimal = int(time_scaled).bit_length() - 1
+    decimal = time_scaled / (1 << pre_decimal) - 1
+
+    value = pre_decimal + decimal
+    return int(32 * value)
+
+
+def native_value_to_time(value: int) -> int:
+    """Convert native LCN value to time.
+
+    Scales the given byte value (0..255) to a time value in milliseconds.
+    Inverse to time_to_native_value().
+
+    :param    int    value:    Duration of timer in native LCN units
+
+    :returns: The duration in milliseconds
+    :rtype: int
+    """
+    if not 0 <= value <= 255:
+        raise ValueError("Value has to be in range 0..255.")
+    pre_decimal = value // 32
+    decimal = value / 32 - pre_decimal
+
+    time_scaled = (1 << pre_decimal) * (decimal + 1)
+
+    time_msec = (time_scaled - 1) * 1000 * 0.03 * 32
+    return int(time_msec)
 
 
 class Var(Enum):
@@ -252,7 +327,7 @@ class Var(Enum):
     S0INPUT4 = auto()  # LCN-BU4LJVarValue
 
     @staticmethod
-    def var_id_to_var(var_id):
+    def var_id_to_var(var_id: int) -> "Var":
         """Translate a given id into a variable type.
 
         :param    int    varId:    The variable id (0..11)
@@ -260,12 +335,12 @@ class Var(Enum):
         :returns: The translated variable enum.
         :rtype: Var
         """
-        if (var_id < 0) or (var_id >= len(Var.variables)):
+        if (var_id < 0) or (var_id >= len(Var.variables)):  # type: ignore
             raise ValueError("Bad var_id.")
-        return Var.variables[var_id]
+        return Var.variables[var_id]  # type: ignore
 
     @staticmethod
-    def set_point_id_to_var(set_point_id):
+    def set_point_id_to_var(set_point_id: int) -> "Var":
         """Translate a given id into a LCN set-point variable type.
 
         :param     int    set_point_id:    Set-point id 0..1
@@ -273,12 +348,12 @@ class Var(Enum):
         :return: The translated var
         :rtype:  Var
         """
-        if (set_point_id < 0) or (set_point_id >= len(Var.set_points)):
+        if (set_point_id < 0) or (set_point_id >= len(Var.set_points)):  # type: ignore
             raise ValueError("Bad set_point_id.")
-        return Var.set_points[set_point_id]
+        return Var.set_points[set_point_id]  # type: ignore
 
     @staticmethod
-    def thrs_id_to_var(register_id, thrs_id):
+    def thrs_id_to_var(register_id: int, thrs_id: int) -> "Var":
         """Translate given ids into a LCN threshold variable type.
 
         :param    int    register_id:    Register id 0..3
@@ -290,15 +365,15 @@ class Var(Enum):
         """
         if (
             (register_id < 0)
-            or (register_id >= len(Var.thresholds))
+            or (register_id >= len(Var.thresholds))  # type: ignore
             or (thrs_id < 0)
             or (thrs_id >= (5 if (register_id == 0) else 4))
         ):
             raise ValueError("Bad register_id and/or thrs_id.")
-        return Var.thresholds[register_id][thrs_id]
+        return Var.thresholds[register_id][thrs_id]  # type: ignore
 
     @staticmethod
-    def s0_id_to_var(s0_id):
+    def s0_id_to_var(s0_id: int) -> "Var":
         """Translate a given id into a LCN S0-input variable type.
 
         :param     int    s0_id:     S0 id 0..3
@@ -306,12 +381,12 @@ class Var(Enum):
         :return:    The translated var
         :rtype:     Var
         """
-        if (s0_id < 0) or (s0_id >= len(Var.s0s)):
+        if (s0_id < 0) or (s0_id >= len(Var.s0s)):  # type: ignore
             raise ValueError("Bad s0_id.")
-        return Var.s0s[s0_id]
+        return Var.s0s[s0_id]  # type: ignore
 
     @staticmethod
-    def to_var_id(var):
+    def to_var_id(var: "Var") -> int:
         """Translate a given variable type into a variable id.
 
         :param     Var    var:    The variable type to translate
@@ -348,7 +423,7 @@ class Var(Enum):
         return var_id
 
     @staticmethod
-    def to_set_point_id(var):
+    def to_set_point_id(var: "Var") -> int:
         """Translate a given variable type into a set-point id.
 
         :param     Var    var:     The variable type to translate
@@ -365,7 +440,7 @@ class Var(Enum):
         return set_point_id
 
     @staticmethod
-    def to_thrs_register_id(var):
+    def to_thrs_register_id(var: "Var") -> int:
         """Translate a given variable type into a threshold register id.
 
         :param    Var    var:    The variable type to translate
@@ -386,7 +461,7 @@ class Var(Enum):
         return thrs_register_id
 
     @staticmethod
-    def to_thrs_id(var):
+    def to_thrs_id(var: "Var") -> int:
         """Translate a given variable type into a threshold id.
 
         :param    Var    var:    The variable type to translate
@@ -409,7 +484,7 @@ class Var(Enum):
         return thrs_id
 
     @staticmethod
-    def to_s0_id(var):
+    def to_s0_id(var: "Var") -> int:
         """Translate a given variable type into an S0-input id.
 
         :param    Var    var:    The variable type to translate
@@ -430,7 +505,7 @@ class Var(Enum):
         return s0_id
 
     @staticmethod
-    def is_lockable_regulator_source(var):
+    def is_lockable_regulator_source(var: "Var") -> bool:
         """Check if the the given variable type is lockable.
 
         :param    Var    var:    The variable type to check
@@ -441,7 +516,7 @@ class Var(Enum):
         return var in [Var.R1VARSETPOINT, Var.R2VARSETPOINT]
 
     @staticmethod
-    def use_lcn_special_values(var):
+    def use_lcn_special_values(var: "Var") -> bool:
         """Check if the given variable type uses special values.
 
         Examples for special values: 'No value yet', 'sensor defective' etc.
@@ -454,7 +529,7 @@ class Var(Enum):
         return var not in [Var.S0INPUT1, Var.S0INPUT2, Var.S0INPUT3, Var.S0INPUT4]
 
     @staticmethod
-    def has_type_in_response(var, sw_age):
+    def has_type_in_response(var: "Var", software_serial: int) -> bool:
         """Module-generation check.
 
         Check if the given variable type would receive a typed response if
@@ -467,7 +542,7 @@ class Var(Enum):
                     otherwise False
         :rtype:    bool
         """
-        if sw_age < 0x170206:
+        if software_serial < 0x170206:
             if var in [
                 Var.VAR1ORTVAR,
                 Var.VAR2ORR1VAR,
@@ -479,7 +554,7 @@ class Var(Enum):
         return True
 
     @staticmethod
-    def is_event_based(var, sw_age):
+    def is_event_based(var: "Var", software_serial: int) -> bool:
         """Module-generation check.
 
         Check if the given variable type automatically sends status-updates
@@ -494,10 +569,10 @@ class Var(Enum):
         """
         if (Var.to_set_point_id(var) != -1) or (Var.to_s0_id(var) != -1):
             return True
-        return sw_age >= 0x170206
+        return software_serial >= 0x170206
 
     @staticmethod
-    def should_poll_status_after_command(var, is2013):
+    def should_poll_status_after_command(var: "Var", is2013: bool) -> bool:
         """Module-generation check.
 
         Check if the target LCN module would automatically send status-updates
@@ -516,7 +591,7 @@ class Var(Enum):
             return False
 
         # Thresholds since 170206 will send status-messages on every change
-        if is2013 and (Var.to_thrs_register_id(var != -1)):
+        if is2013 and (Var.to_thrs_register_id(var) != -1):
             return False
 
         # Others:
@@ -529,7 +604,9 @@ class Var(Enum):
         return True
 
     @staticmethod
-    def should_poll_status_after_regulator_lock(sw_age, lock_state):
+    def should_poll_status_after_regulator_lock(
+        software_serial: int, lock_state: int
+    ) -> bool:
         """Module-generation check.
 
         Check if the target LCN module would automatically send status-updates
@@ -544,11 +621,11 @@ class Var(Enum):
         """
         # LCN modules before 170206 will send an automatic status-message for
         # "lock", but not for "unlock"
-        return (not lock_state) and (sw_age < 0x170206)
+        return (not lock_state) and (software_serial < 0x170206)
 
 
 # Helper list to get var by numeric id.
-Var.variables = [
+Var.variables = [  # type: ignore
     Var.VAR1ORTVAR,
     Var.VAR2ORR1VAR,
     Var.VAR3ORR2VAR,
@@ -564,10 +641,10 @@ Var.variables = [
 ]
 
 # Helper list to get set-point var by numeric id.
-Var.set_points = [Var.R1VARSETPOINT, Var.R2VARSETPOINT]
+Var.set_points = [Var.R1VARSETPOINT, Var.R2VARSETPOINT]  # type: ignore
 
 # Helper list to get threshold var by numeric id.
-Var.thresholds = [
+Var.thresholds = [  # type: ignore
     [Var.THRS1, Var.THRS2, Var.THRS3, Var.THRS4, Var.THRS5],
     [Var.THRS2_1, Var.THRS2_2, Var.THRS2_3, Var.THRS2_4],
     [Var.THRS3_1, Var.THRS3_2, Var.THRS3_3, Var.THRS3_4],
@@ -575,16 +652,21 @@ Var.thresholds = [
 ]
 
 # Helper list to get S0-input var by numeric id.
-Var.s0s = [Var.S0INPUT1, Var.S0INPUT2, Var.S0INPUT3, Var.S0INPUT4]
+Var.s0s = [  # type: ignore
+    Var.S0INPUT1,
+    Var.S0INPUT2,
+    Var.S0INPUT3,
+    Var.S0INPUT4,
+]
 
 
 class VarUnit(Enum):
     """Measurement units used with LCN variables."""
 
     NATIVE = ""  # LCN internal representation (0 = -100C for absolute values)
-    CELSIUS = u"\u00b0C"
-    KELVIN = u"\u00b0K"
-    FAHRENHEIT = u"\u00b0F"
+    CELSIUS = "\u00b0C"
+    KELVIN = "\u00b0K"
+    FAHRENHEIT = "\u00b0F"
     LUX_T = "Lux_T"
     LUX_I = "Lux_I"
     METERPERSECOND = "m/s"  # Used for LCN-WIH wind speed
@@ -592,10 +674,10 @@ class VarUnit(Enum):
     PPM = "ppm"  # Used by CO2 sensor
     VOLT = "V"
     AMPERE = "A"
-    DEGREE = u"\u00b0"  # Used for angles,
+    DEGREE = "\u00b0"  # Used for angles,
 
     @staticmethod
-    def parse(unit):
+    def parse(unit: str) -> "VarUnit":
         """Parse the given unit string and return VarUnit.
 
         :param    str    unit:    The input unit
@@ -603,11 +685,11 @@ class VarUnit(Enum):
         unit = unit.upper()
         if unit in ["", "NATIVE", "LCN"]:
             var_unit = VarUnit.NATIVE
-        elif unit in ["CELSIUS", u"\u00b0CELSIUS", u"\u00b0C"]:
+        elif unit in ["CELSIUS", "\u00b0CELSIUS", "\u00b0C"]:
             var_unit = VarUnit.CELSIUS
-        elif unit in ["KELVIN", u"\u00b0KELVIN", u"\u00b0K", "K"]:
+        elif unit in ["KELVIN", "\u00b0KELVIN", "\u00b0K", "K"]:
             var_unit = VarUnit.KELVIN
-        elif unit in ["FAHRENHEIT", u"\u00b0FAHRENHEIT", u"\u00b0F"]:
+        elif unit in ["FAHRENHEIT", "\u00b0FAHRENHEIT", "\u00b0F"]:
             var_unit = VarUnit.FAHRENHEIT
         elif unit in ["LUX_T", "LX_T"]:
             var_unit = VarUnit.LUX_T
@@ -623,7 +705,7 @@ class VarUnit(Enum):
             var_unit = VarUnit.VOLT
         elif unit in ["AMPERE", "AMP", "A"]:
             var_unit = VarUnit.AMPERE
-        elif unit in ["DEGREE", u"\u00b0"]:
+        elif unit in ["DEGREE", "\u00b0"]:
             var_unit = VarUnit.DEGREE
         else:
             raise ValueError("Bad input unit.")
@@ -641,19 +723,29 @@ class VarValue:
     :param    int    native_value:    The native value
     """
 
-    def __init__(self, native_value):
+    def __init__(self, native_value: int) -> None:
         """Construct with native LCN value."""
         self.native_value = native_value
 
-    def is_locked_regulator(self):
+    def __eq__(self, other: object) -> bool:
+        """Return if instance equals the given object."""
+        if isinstance(other, VarValue):
+            return self.native_value == other.native_value
+        return False
+
+    def __hash__(self) -> int:
+        """Calculate the instance hash value."""
+        return self.native_value.__hash__()
+
+    def is_locked_regulator(self) -> bool:
         """Return if regulator is locked."""
         return (self.native_value & 0x8000) != 0
 
     @staticmethod
-    def from_var_unit(v, unit, is_abs):
+    def from_var_unit(value: float, unit: VarUnit, is_abs: bool) -> "VarValue":
         """Create a variable value from any input.
 
-        :param    float    v:       The input value
+        :param    float    value:   The input value
         :param    VarUnit  unit:    The input value's unit
         :param    bool     is_abs:  True for absolute values (relative values
                                     are used to add/subtract from other
@@ -663,49 +755,49 @@ class VarValue:
         :rtype:    VarValue
         """
         if unit == VarUnit.NATIVE:
-            var_value = VarValue.from_native(int(v))
+            var_value = VarValue.from_native(int(value))
         elif unit == VarUnit.CELSIUS:
-            var_value = VarValue.from_celsius(v, is_abs)
+            var_value = VarValue.from_celsius(value, is_abs)
         elif unit == VarUnit.KELVIN:
-            var_value = VarValue.from_kelvin(v, is_abs)
+            var_value = VarValue.from_kelvin(value, is_abs)
         elif unit == VarUnit.FAHRENHEIT:
-            var_value = VarValue.from_fahrenheit(v, is_abs)
+            var_value = VarValue.from_fahrenheit(value, is_abs)
         elif unit == VarUnit.LUX_T:
-            var_value = VarValue.from_lux_t(v)
+            var_value = VarValue.from_lux_t(value)
         elif unit == VarUnit.LUX_I:
-            var_value = VarValue.from_lux_i(v)
+            var_value = VarValue.from_lux_i(value)
         elif unit == VarUnit.METERPERSECOND:
-            var_value = VarValue.from_meters_per_second(v)
+            var_value = VarValue.from_meters_per_second(value)
         elif unit == VarUnit.PERCENT:
-            var_value = VarValue.from_percent(v)
+            var_value = VarValue.from_percent(value)
         elif unit == VarUnit.PPM:
-            var_value = VarValue.from_ppm(v)
+            var_value = VarValue.from_ppm(value)
         elif unit == VarUnit.VOLT:
-            var_value = VarValue.from_volt(v)
+            var_value = VarValue.from_volt(value)
         elif unit == VarUnit.AMPERE:
-            var_value = VarValue.from_kelvin(v)
+            var_value = VarValue.from_kelvin(value)
         elif unit == VarUnit.DEGREE:
-            var_value = VarValue.from_degree(v)
+            var_value = VarValue.from_degree(value)
         else:
             raise ValueError("Wrong unit.")
         return var_value
 
     @staticmethod
-    def from_native(n):
+    def from_native(value: int) -> "VarValue":
         """Create a variable value from native input.
 
-        :param    int    n:    The input value
+        :param    int    value:    The input value
 
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(n)
+        return VarValue(value)
 
     @staticmethod
-    def from_celsius(c, is_abs=True):
+    def from_celsius(value: float, is_abs: bool = True) -> "VarValue":
         """Create a variable value from Celsius input.
 
-        :param    float    c:        The input value
+        :param    float    value:    The input value
         :param    bool     is_abs:   True for absolute values (relative values
                                      are used to add/subtract from other
                                      VarValues), otherwise False
@@ -713,14 +805,14 @@ class VarValue:
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        n = int(round(c * 10))
-        return VarValue(n + 1000 if is_abs else n)
+        number = int(round(value * 10))
+        return VarValue(number + 1000 if is_abs else number)
 
     @staticmethod
-    def from_kelvin(k, is_abs=True):
+    def from_kelvin(value: float, is_abs: bool = True) -> "VarValue":
         """Create a variable value from Kelvin input.
 
-        :param    float    k:        The input value
+        :param    float    value:    The input value
         :param    bool     is_abs:   True for absolute values (relative values
                                      are used to add/subtract from other
                                      VarValues), otherwise False
@@ -729,16 +821,16 @@ class VarValue:
         :rtype:    VarValue
         """
         if is_abs:
-            k -= 273.15
+            value -= 273.15
 
-        n = int(round(k * 10))
-        return VarValue(n + 1000 if is_abs else n)
+        number = int(round(value * 10))
+        return VarValue(number + 1000 if is_abs else number)
 
     @staticmethod
-    def from_fahrenheit(f, is_abs=True):
+    def from_fahrenheit(value: float, is_abs: bool = True) -> "VarValue":
         """Create a variable value from Fahrenheit input.
 
-        :param    float    f:        The input value
+        :param    float    value:    The input value
         :param    bool     is_abs:   True for absolute values (relative values
                                      are used to add/subtract from other
                                      VarValues), otherwise False
@@ -747,13 +839,13 @@ class VarValue:
         :rtype:    VarValue
         """
         if is_abs:
-            f -= 32
+            value -= 32
 
-        n = int(round(f / 0.18))
-        return VarValue(n + 1000 if is_abs else n)
+        number = int(round(value / 0.18))
+        return VarValue(number + 1000 if is_abs else number)
 
     @staticmethod
-    def from_lux_t(lux):
+    def from_lux_t(lux: float) -> "VarValue":
         """Create a variable value from lx input.
 
         Target must be connected to T-port.
@@ -766,7 +858,7 @@ class VarValue:
         return VarValue(int(round(math.log(lux) - 1.689646994) / 0.010380664))
 
     @staticmethod
-    def from_lux_i(lux):
+    def from_lux_i(lux: float) -> "VarValue":
         """Create a variable value from lx input.
 
         Target must be connected to I-port.
@@ -779,69 +871,69 @@ class VarValue:
         return VarValue(int(round(math.log(lux) * 100)))
 
     @staticmethod
-    def from_percent(p):
+    def from_percent(value: float) -> "VarValue":
         """Create a variable value from % input.
 
-        :param    float    p:    The input value
+        :param    float    value:    The input value
 
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(int(round(p)))
+        return VarValue(int(round(value)))
 
     @staticmethod
-    def from_ppm(p):
+    def from_ppm(value: float) -> "VarValue":
         """Create a variable value from ppm input.
 
         Used for CO2 sensors.
 
-        :param    float    p:   The input value
+        :param    float    value:   The input value
 
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(int(round(p)))
+        return VarValue(int(round(value)))
 
     @staticmethod
-    def from_meters_per_second(ms):
+    def from_meters_per_second(value: float) -> "VarValue":
         """Create a variable value from m/s input.
 
         Used for LCN-WIH wind speed.
 
-        :param    float    ms:    The input value
+        :param    float    value:    The input value
 
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(int(round(ms * 10)))
+        return VarValue(int(round(value * 10)))
 
     @staticmethod
-    def from_volt(v):
+    def from_volt(value: float) -> "VarValue":
         """Create a variable value from V input.
 
-        :param    float    v:    The input value
+        :param    float    value:    The input value
 
         :return:    The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(int(round(v * 400)))
+        return VarValue(int(round(value * 400)))
 
     @staticmethod
-    def from_ampere(a):
+    def from_ampere(value: float) -> "VarValue":
         """Create a variable value from A input.
 
-        :param    float    a:    The input value
+        :param    float    value:    The input value
 
         :return: The variable value (never null)
         :rtype:    VarValue
         """
-        return VarValue(int(round(a * 100)))
+        return VarValue(int(round(value * 100)))
 
     @staticmethod
-    def from_degree(d, is_abs=True):
+    def from_degree(value: float, is_abs: bool = True) -> "VarValue":
         """Create a variable value from degree (angle) input.
 
-        :param    float    d:        The input value
+        :param    float    value:        The input value
         :param    bool     is_abs:   True for absolute values (relative values
                                      are used to add/subtract from other
                                      VarValues), otherwise False
@@ -849,53 +941,53 @@ class VarValue:
         :return:    The variable value (never null)
         :rtype:     VarValue
         """
-        n = int(round(d * 10))
-        return VarValue(n + 1000 if is_abs else n)
+        number = int(round(value * 10))
+        return VarValue(number + 1000 if is_abs else number)
 
-    def to_var_unit(self, unit, is_lockable_regulator_source=False):
+    def to_var_unit(
+        self, unit: VarUnit, is_lockable_regulator_source: bool = False
+    ) -> Union[int, float]:
         """Convert the given unit to a VarValue.
 
         :param    VarUnit    unit:    The variable unit
         :param    bool       is_lockable_regulator_source:  Is lockable source
 
         :return:    The variable value
-        :rtype:     VarValue
+        :rtype:     Union[int,float]
         """
-        v = VarValue(
+        var_value = VarValue(
             self.native_value & 0x7FFF
             if is_lockable_regulator_source
             else self.native_value
         )
 
         if unit == VarUnit.NATIVE:
-            var_value = v.to_native()
-        elif unit == VarUnit.CELSIUS:
-            var_value = v.to_celsius()
-        elif unit == VarUnit.KELVIN:
-            var_value = v.to_kelvin()
-        elif unit == VarUnit.FAHRENHEIT:
-            var_value = v.to_fahrenheit()
-        elif unit == VarUnit.LUX_T:
-            var_value = v.to_lux_t()
-        elif unit == VarUnit.LUX_I:
-            var_value = v.to_lux_i()
-        elif unit == VarUnit.METERPERSECOND:
-            var_value = v.to_meters_per_second()
-        elif unit == VarUnit.PERCENT:
-            var_value = v.to_percent()
-        elif unit == VarUnit.PPM:
-            var_value = v.to_ppm()
-        elif unit == VarUnit.VOLT:
-            var_value = v.to_volt()
-        elif unit == VarUnit.AMPERE:
-            var_value = v.to_ampere()
-        elif unit == VarUnit.DEGREE:
-            var_value = v.to_degree()
-        else:
-            raise ValueError("Wrong unit.")
-        return var_value
+            return var_value.to_native()
+        if unit == VarUnit.CELSIUS:
+            return var_value.to_celsius()
+        if unit == VarUnit.KELVIN:
+            return var_value.to_kelvin()
+        if unit == VarUnit.FAHRENHEIT:
+            return var_value.to_fahrenheit()
+        if unit == VarUnit.LUX_T:
+            return var_value.to_lux_t()
+        if unit == VarUnit.LUX_I:
+            return var_value.to_lux_i()
+        if unit == VarUnit.METERPERSECOND:
+            return var_value.to_meters_per_second()
+        if unit == VarUnit.PERCENT:
+            return var_value.to_percent()
+        if unit == VarUnit.PPM:
+            return var_value.to_ppm()
+        if unit == VarUnit.VOLT:
+            return var_value.to_volt()
+        if unit == VarUnit.AMPERE:
+            return var_value.to_ampere()
+        if unit == VarUnit.DEGREE:
+            return var_value.to_degree()
+        raise ValueError("Wrong unit.")
 
-    def to_native(self):
+    def to_native(self) -> int:
         """Convert to native value.
 
         :return:    The converted value
@@ -903,7 +995,7 @@ class VarValue:
         """
         return self.native_value
 
-    def to_celsius(self):
+    def to_celsius(self) -> float:
         """Convert to Celsius value.
 
         :return:    The converted value
@@ -911,7 +1003,7 @@ class VarValue:
         """
         return (self.native_value - 1000) / 10.0
 
-    def to_kelvin(self):
+    def to_kelvin(self) -> float:
         """Convert to Kelvin value.
 
         :return:    The converted value
@@ -919,7 +1011,7 @@ class VarValue:
         """
         return (self.native_value - 1000) / 10.0 + 273.15
 
-    def to_fahrenheit(self):
+    def to_fahrenheit(self) -> float:
         """Convert to Fahrenheit value.
 
         :return:    The converted value
@@ -927,7 +1019,7 @@ class VarValue:
         """
         return (self.native_value - 1000) * 0.18 + 32.0
 
-    def to_lux_t(self):
+    def to_lux_t(self) -> float:
         """Convert to lx value.
 
         Source must be connected to T-port.
@@ -937,7 +1029,7 @@ class VarValue:
         """
         return math.exp(0.010380664 * self.native_value + 1.689646994)
 
-    def to_lux_i(self):
+    def to_lux_i(self) -> float:
         """Convert to lx value.
 
         Source must be connected to I-port.
@@ -947,7 +1039,7 @@ class VarValue:
         """
         return math.exp(self.native_value / 100)
 
-    def to_percent(self):
+    def to_percent(self) -> int:
         """Convert to % value.
 
         :return:    The converted value
@@ -955,7 +1047,7 @@ class VarValue:
         """
         return self.native_value
 
-    def to_ppm(self):
+    def to_ppm(self) -> int:
         """Convert to ppm value.
 
         :return:    The converted value
@@ -963,7 +1055,7 @@ class VarValue:
         """
         return self.native_value
 
-    def to_meters_per_second(self):
+    def to_meters_per_second(self) -> float:
         """Convert to m/s value.
 
         :return:    The converted value
@@ -971,7 +1063,7 @@ class VarValue:
         """
         return self.native_value / 10.0
 
-    def to_volt(self):
+    def to_volt(self) -> float:
         """Convert to V value.
 
         :return:    The converted value
@@ -979,7 +1071,7 @@ class VarValue:
         """
         return self.native_value / 400.0
 
-    def to_ampere(self):
+    def to_ampere(self) -> float:
         """Convert to A value.
 
         :return:    The converted value
@@ -987,7 +1079,7 @@ class VarValue:
         """
         return self.native_value / 100.0
 
-    def to_degree(self):
+    def to_degree(self) -> float:
         """Convert to degree value.
 
         :return:    The converted value
@@ -996,8 +1088,11 @@ class VarValue:
         return (self.native_value - 1000) / 10.0
 
     def to_var_unit_string(
-        self, unit, is_lockable_regulator_source=False, use_lcn_special_values=False
-    ):
+        self,
+        unit: VarUnit,
+        is_lockable_regulator_source: bool = False,
+        use_lcn_special_values: bool = False,
+    ) -> str:
         """Convert the given unit into a string representation.
 
         :param    VarUnit    unit:    The input unit
@@ -1024,41 +1119,41 @@ class VarValue:
                 else self.native_value
             )
             if unit == VarUnit.NATIVE:
-                ret = "{:.0f}".format(var.to_native())
+                ret = f"{var.to_native():.0f}"
             elif unit == VarUnit.CELSIUS:
-                ret = "{:.01f}".format(var.to_celsius())
+                ret = f"{var.to_celsius():.01f}"
             elif unit == VarUnit.KELVIN:
-                ret = "{:.01f}".format(var.to_kelvin())
+                ret = f"{var.to_kelvin():.01f}"
             elif unit == VarUnit.FAHRENHEIT:
-                ret = "{:.01f}".format(var.to_fahrenheit())
+                ret = f"{var.to_fahrenheit():.01f}"
             elif unit == VarUnit.LUX_T:
                 if var.to_native() > 1152:  # Max. value the HW can do
                     ret = "---"
                 else:
-                    ret = "{:.0f}".format(var.to_lux_t())
+                    ret = f"{var.to_lux_t():.0f}"
             elif unit == VarUnit.LUX_I:
                 if var.to_native() > 1152:  # Max. value the HW can do
                     ret = "---"
                 else:
-                    ret = "{:.0f}".format(var.to_lux_i())
+                    ret = f"{var.to_lux_i():.0f}"
             elif unit == VarUnit.METERPERSECOND:
-                ret = "{:.0f}".format(var.to_meters_per_second())
+                ret = f"{var.to_meters_per_second():.0f}"
             elif unit == VarUnit.PERCENT:
-                ret = "{:.0f}".format(var.to_percent())
+                ret = f"{var.to_percent():.0f}"
             elif unit == VarUnit.PPM:
-                ret = "{:.0f}".format(var.to_ppm())
+                ret = f"{var.to_ppm():.0f}"
             elif unit == VarUnit.VOLT:
-                ret = "{:.0f}".format(var.to_volt())
+                ret = f"{var.to_volt():.0f}"
             elif unit == VarUnit.AMPERE:
-                ret = "{:.0f}".format(var.to_ampere())
+                ret = f"{var.to_ampere():.0f}"
             elif unit == VarUnit.DEGREE:
-                ret = "{:.0f}".format(var.to_degree())
+                ret = f"{var.to_degree():.0f}"
             else:
                 raise ValueError("Wrong unit.")
 
         # handle locked regulators
         if is_lockable_regulator_source and self.is_locked_regulator():
-            ret = "({:s})".format(ret)
+            ret = f"({ret:s})"
 
         return ret
 
@@ -1075,9 +1170,9 @@ class LedStatus(Enum):
 class LogicOpStatus(Enum):
     """Possible states for LCN logic-operations."""
 
-    NOT = "N"
-    OR = "T"  # Note: Actually not correct since AND won't be OR also
-    AND = "V"
+    NONE = "N"
+    SOME = "T"  # Note: Actually not correct since AND won't be OR also
+    ALL = "V"
 
 
 class TimeUnit(Enum):
@@ -1089,7 +1184,7 @@ class TimeUnit(Enum):
     DAYS = "D"
 
     @staticmethod
-    def parse(time_unit):
+    def parse(unit: str) -> "TimeUnit":
         """Parse the given time_unit into a time unit.
 
         It supports several alternative terms.
@@ -1099,18 +1194,18 @@ class TimeUnit(Enum):
         :return:    TimeUnit enum
         :rtype:    TimeUnit
         """
-        time_unit = time_unit.upper()
-        if time_unit in ["SECONDS", "SECOND", "SEC", "S"]:
-            tu = TimeUnit.SECONDS
-        elif time_unit in ["MINUTES", "MINUTE", "MIN", "M"]:
-            tu = TimeUnit.MINUTES
-        elif time_unit in ["HOURS", "HOUR", "H"]:
-            tu = TimeUnit.HOURS
-        elif time_unit in ["DAYS", "DAY", "D"]:
-            tu = TimeUnit.DAYS
+        unit = unit.upper()
+        if unit in ["SECONDS", "SECOND", "SEC", "S"]:
+            time_unit = TimeUnit.SECONDS
+        elif unit in ["MINUTES", "MINUTE", "MIN", "M"]:
+            time_unit = TimeUnit.MINUTES
+        elif unit in ["HOURS", "HOUR", "H"]:
+            time_unit = TimeUnit.HOURS
+        elif unit in ["DAYS", "DAY", "D"]:
+            time_unit = TimeUnit.DAYS
         else:
             raise ValueError("Bad time unit input.")
-        return tu
+        return time_unit
 
 
 class RelayStateModifier(Enum):
@@ -1132,7 +1227,7 @@ class MotorStateModifier(Enum):
     DOWN = "D"
     STOP = "S"
     TOGGLEONOFF = "T"  # toggle on/off
-    TOGGLEDIR = "D"  # toggle direction
+    TOGGLEDIR = "R"  # toggle direction
     CYCLE = "C"  # up, stop, down, stop, ...
     NOCHANGE = "-"
 
@@ -1173,12 +1268,109 @@ class KeyLockStateModifier(Enum):
     NOCHANGE = "-"
 
 
-default_connection_settings = {
+class BeepSound(Enum):
+    """Beep sounds supported by LCN modules."""
+
+    NORMAL = "N"
+    SPECIAL = "S"
+
+
+HARDWARE_DESCRIPTIONS = dict(
+    [
+        (-1, "UnknownModuleType"),
+        (1, "LCN-SW1.0"),
+        (2, "LCN-SW1.1"),
+        (3, "LCN-UP1.0"),
+        (4, "LCN-UP2"),
+        (5, "LCN-SW2"),
+        (6, "LCN-UP-Profi1-Plus"),
+        (7, "LCN-DI12"),
+        (8, "LCN-HU"),
+        (9, "LCN-SH"),
+        (10, "LCN-UP2"),
+        (11, "LCN-UPP"),
+        (12, "LCN-SK"),
+        (14, "LCN-LD"),
+        (15, "LCN-SH-Plus"),
+        (17, "LCN-UPS"),
+        (18, "LCN_UPS24V"),
+        (19, "LCN-GTM"),
+        (20, "LCN-SHS"),
+        (21, "LCN-ESD"),
+        (22, "LCN-EB2"),
+        (23, "LCN-MRS"),
+        (24, "LCN-EB11"),
+        (25, "LCN-UMR"),
+        (26, "LCN-UPU"),
+        (27, "LCN-UMR24V"),
+        (28, "LCN-SHD"),
+        (29, "LCN-SHU"),
+        (30, "LCN-SR6"),
+    ]
+)
+
+
+class HardwareType(Enum):
+    """Hardware types as returned by serial number request."""
+
+    UNKNOWN = -1
+    SW1_0 = 1
+    SW1_1 = 2
+    UP1_0 = 3
+    UP2 = 4
+    SW2 = 5
+    UP_PROFI1_PLUS = 6
+    DI12 = 7
+    HU = 8
+    SH = 9
+    UPP = 11
+    SK = 12
+    LD = 14
+    SH_PLUS = 15
+    UPS = 17
+    UPS24V = 18
+    GTM = 19
+    SHS = 20
+    ESD = 21
+    EB2 = 22
+    MRS = 23
+    EB11 = 24
+    UMR = 25
+    UPU = 26
+    UMR24V = 27
+    SHD = 28
+    SHU = 29
+    SR6 = 30
+
+    @property
+    def identifier(self) -> Any:
+        """Get the LCN hardware identifier."""
+        return self.value
+
+    @property
+    def description(self) -> str:
+        """Get the LCN hardware name."""
+        return HARDWARE_DESCRIPTIONS[self.value]
+
+
+@no_type_check
+def hw_type_new(cls, value):
+    """Replace Hardwaretype.__new__."""
+    if value == 10:
+        value = 4
+    return super(HardwareType, cls).__new__(cls, value)
+
+
+setattr(HardwareType, "__new__", hw_type_new)
+
+
+default_connection_settings: Dict[str, Any] = {
     "NUM_TRIES": 3,  # Total number of request to sent before going into
     # failed-state.
     "SK_NUM_TRIES": 3,  # Total number of segment coupler scan tries
     "DIM_MODE": OutputPortDimMode.STEPS50,
     "PING_TIMEOUT": 600000,  # The default timeout for pings sent to PCHK.
+    "DEFAULT_TIMEOUT_MSEC": 3500,  # Default timeout for send command retries
     "MAX_STATUS_EVENTBASED_VALUEAGE_MSEC": 600000,  # Poll interval for
     # status values that
     # automatically send
@@ -1193,4 +1385,3 @@ default_connection_settings = {
     # potentially changed
     # that status.
 }
-
