@@ -66,8 +66,6 @@ class AbstractConnection:
             software_serial = -1
         self._software_serial: int = software_serial
 
-        self.input_callbacks: List[Callable[[inputs.Input], None]] = []
-
     @property
     def seg_id(self) -> int:
         """Get the segment id."""
@@ -137,29 +135,6 @@ class AbstractConnection:
             return await self.conn.send_command(header + pck)
         return await self.conn.send_command(header.encode() + pck)
 
-    # ##
-    # ## Methods for handling input objects
-    # ##
-
-    async def async_process_input(self, inp: inputs.Input) -> None:
-        """Is called by input object's process method.
-
-        Method to handle incoming commands for this specific module (status,
-        toggle_output, switch_relays, ...)
-        """
-        for input_callback in self.input_callbacks:
-            input_callback(inp)
-
-    def register_for_inputs(
-        self, callback: Callable[[inputs.Input], None]
-    ) -> Callable[..., None]:
-        """Register a function for callback on PCK message received.
-
-        Returns a function to unregister the callback.
-        """
-        self.input_callbacks.append(callback)
-        return lambda callback=callback: self.input_callbacks.remove(callback)
-
     # ##
     # ## Methods for sending PCK commands
     # ##
@@ -822,6 +797,8 @@ class ModuleConnection(AbstractConnection):
         self.activate_status_requests = activate_status_requests
         self.has_s0_enabled = has_s0_enabled
 
+        self.input_callbacks: List[Callable[[inputs.Input], None]] = []
+
         # List of queued acknowledge codes from the LCN modules.
         self.acknowledges: "asyncio.Queue[int]" = asyncio.Queue()
 
@@ -950,6 +927,20 @@ class ModuleConnection(AbstractConnection):
         """Get the activation status for S0 variables."""
         return self.has_s0_enabled
 
+    # ##
+    # ## Methods for handling input objects
+    # ##
+
+    def register_for_inputs(
+        self, callback: Callable[[inputs.Input], None]
+    ) -> Callable[..., None]:
+        """Register a function for callback on PCK message received.
+
+        Returns a function to unregister the callback.
+        """
+        self.input_callbacks.append(callback)
+        return lambda callback=callback: self.input_callbacks.remove(callback)
+
     async def async_process_input(self, inp: inputs.Input) -> None:
         """Is called by input object's process method.
 
@@ -964,7 +955,33 @@ class ModuleConnection(AbstractConnection):
         if isinstance(inp, inputs.ModStatusVar):
             inp = self.status_requests_handler.preprocess_modstatusvar(inp)
 
-        await super().async_process_input(inp)
+        for input_callback in self.input_callbacks:
+            input_callback(inp)
+
+    def dump_details(self) -> Dict[str, Any]:
+        """Dump detailed information about this module."""
+        is_local_segment = (
+            self.addr.seg_id == 0 or self.addr.seg_id == self.conn.local_seg_id
+        )
+        return {
+            "segment": self.addr.seg_id,
+            "address": self.addr.addr_id,
+            "is_local_segment": is_local_segment,
+            "serials": {
+                "hardware_serial": f"{self.hardware_serial:10X}",
+                "manu": f"{self.manu:02X}",
+                "software_serial": f"{self.software_serial:06X}",
+                "hardware_type": f"{self.hardware_type.value:d}",
+                "hardware_name": self.hardware_type.description,
+            },
+            "name": self.name,
+            "comment": self.comment,
+            "oem_text": self.oem_text,
+            "groups": {
+                "static": sorted(addr.addr_id for addr in self.static_groups),
+                "dynamic": sorted(addr.addr_id for addr in self.dynamic_groups),
+            },
+        }
 
     # ##
     # ## Requests
