@@ -1,9 +1,13 @@
-"""Copyright (c) 2006-2018 by the respective copyright holders.
+"""Copyright (c) 2006-2020 by the respective copyright holders.
 
-All rights reserved. This program and the accompanying materials
-are made available under the terms of the Eclipse Public License v1.0
-which accompanies this distribution, and is available at
-http://www.eclipse.org/legal/epl-v10.html
+See the NOTICE file(s) distributed with this work for additional
+information.
+
+This program and the accompanying materials are made available under the
+terms of the Eclipse Public License 2.0 which is available at
+http://www.eclipse.org/legal/epl-2.0
+
+SPDX-License-Identifier: EPL-2.0
 
 Contributors:
   Andre Lengwenus - port to Python and further improvements
@@ -11,8 +15,10 @@ Contributors:
 """
 
 import re
+from typing import List, Optional, Sequence
 
 from pypck import lcn_defs
+from pypck.lcn_addr import LcnAddr
 
 
 class PckParser:
@@ -40,6 +46,9 @@ class PckParser:
     # LCN-PK/PKU is disconnected.
     LCNCONNSTATE_DISCONNECTED = "$io:#LCN:disconnected"
 
+    # Decimal mode set
+    DEC_MODE_SET = "(dec-mode)"
+
     # License Error
     LICENSE_ERROR = "$err:(license?)"
 
@@ -59,24 +68,33 @@ class PckParser:
 
     # Pattern to parse serial number and firmware date responses.
     PATTERN_SN = re.compile(
-        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3}).SN(?P<sn>[0-9|A-F]{10})"
-        r"(?P<manu>[0-9|A-F]{2})FW(?P<sw_age>[0-9|A-F]{6})HW(?P<hw_type>\d+)"
+        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SN(?P<hardware_serial>[0-9|A-F]{10})"
+        r"(?P<manu>[0-9|A-F]{2})FW(?P<software_serial>[0-9|A-F]{6})"
+        r"HW(?P<hardware_type>\d+)"
     )
 
     # Pattern to parse module name and comment
     PATTERN_NAME_COMMENT = re.compile(
-        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3}).(?P<command>[NKO])"
+        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.(?P<command>[NKO])"
         r"(?P<block_id>\d)(?P<text>.{0,12})"
     )
 
+    # Pattern to parse the static and dynamic group membership status
+    PATTERN_STATUS_GROUPS = re.compile(
+        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.G(?P<kind>[DP])(?P<max_groups>\d{3})"
+        r"(?:(?P<g1>\d{3}))?(?:(?P<g2>\d{3}))?(?:(?P<g3>\d{3}))?(?:(?P<g4>\d{3}))?"
+        r"(?:(?P<g5>\d{3}))?(?:(?P<g6>\d{3}))?(?:(?P<g7>\d{3}))?(?:(?P<g8>\d{3}))?"
+        r"(?:(?P<g9>\d{3}))?(?:(?P<g10>\d{3}))?(?:(?P<g11>\d{3}))?(?:(?P<g12>\d{3}))?"
+    )
+
     # Pattern to parse output-port status responses in percent.
     PATTERN_STATUS_OUTPUT_PERCENT = re.compile(
-        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})A(?P<output_id>\d)" r"(?P<percent>\d+)"
+        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})A(?P<output_id>\d)(?P<percent>\d+)"
     )
 
     # Pattern to parse output-port status responses in native format (0..200).
     PATTERN_STATUS_OUTPUT_NATIVE = re.compile(
-        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})O(?P<output_id>\d)" r"(?P<value>\d+)"
+        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})O(?P<output_id>\d)(?P<value>\d+)"
     )
 
     # Pattern to parse relays status responses.
@@ -136,8 +154,30 @@ class PckParser:
         r"(?P<table1>\d{3})(?P<table2>\d{3})((?P<table3>\d{3}))?"
     )
 
+    # Pattern to parse scene output status messages.
+    PATTERN_STATUS_SCENE_OUTPUTS = re.compile(
+        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SZ(?P<scene_id>\d{3})"
+        r"(?P<output1>\d{3})(?P<ramp1>\d{3})(?P<output2>\d{3})(?P<ramp2>\d{3})"
+        r"(?P<output3>\d{3})(?P<ramp3>\d{3})(?P<output4>\d{3})(?P<ramp4>\d{3})"
+    )
+
+    # Pattern to parse send command to host messages.
+    PATTERN_SEND_COMMAND_HOST = re.compile(
+        r"\+M004(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SKH"
+        r"(?P<p1>\d{3})(?P<p2>\d{3})"
+        r"(?:(?P<p3>\d{3})(?P<p4>\d{3})(?P<p5>\d{3})(?P<p6>\d{3}))?"
+        r"(?:(?P<p7>\d{3})(?P<p8>\d{3})(?P<p9>\d{3})(?P<p10>\d{3})"
+        r"(?P<p11>\d{3})(?P<p12>\d{3})(?P<p13>\d{3})(?P<p14>\d{3}))?"
+    )
+
+    # Pattern to parse send key to host messages.
+    PATTERN_SEND_KEYS_HOST = re.compile(
+        r"\+M004(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.STH"
+        r"(?P<actions>\d{3})(?P<keys>\d{3})"
+    )
+
     @staticmethod
-    def get_boolean_value(input_byte):
+    def get_boolean_value(input_byte: int) -> List[bool]:
         """Get boolean representation for the given byte.
 
         :param    int    input_byte:    Input byte as int8.
@@ -167,11 +207,10 @@ class PckGenerator:
     TABLE_NAMES = ["A", "B", "C", "D"]
 
     @staticmethod
-    def ping(counter):
+    def ping(counter: int) -> str:
         """Generate a keep-alive.
 
         LCN-PCHK will close the connection if it does not receive any commands
-        from an open :class:`~pypck.connection.PchkConnectionManager` for a
         specific period (10 minutes by default).
 
         :param    int    counter:    The current ping's id (optional, but
@@ -179,10 +218,17 @@ class PckGenerator:
         :return:    The PCK command as text
         :rtype:    str
         """
-        return "^ping{:d}".format(counter)
+        return f"^ping{counter:d}"
 
     @staticmethod
-    def set_operation_mode(dim_mode, status_mode):
+    def set_dec_mode() -> str:
+        """Generate PCK command to set used number system to decimal."""
+        return "!CHD"
+
+    @staticmethod
+    def set_operation_mode(
+        dim_mode: lcn_defs.OutputPortDimMode, status_mode: lcn_defs.OutputPortStatusMode
+    ) -> str:
         """Generate a PCK command to set the connection's operation mode.
 
         This influences how output-port commands and status are interpreted
@@ -201,11 +247,13 @@ class PckGenerator:
         )
 
     @staticmethod
-    def generate_address_header(addr, local_seg_id, wants_ack):
+    def generate_address_header(
+        addr: LcnAddr, local_seg_id: int, wants_ack: bool
+    ) -> str:
         """Generate a PCK command address header.
 
         :param    addr:   The module's/group's address
-        :type     addr:   :class:`~LcnAddrMod` or :class:`~LcnAddrGrp`
+        :type     addr:   :class:`~LcnAddr`
         :param    int     local_seg_id:    The local segment id
         :param    bool    wants_ack:      Is an acknowledge requested.
 
@@ -213,14 +261,14 @@ class PckGenerator:
         :rtype:     str
         """
         return ">{:s}{:03d}{:03d}{:s}".format(
-            "G" if addr.is_group() else "M",
+            "G" if addr.is_group else "M",
             addr.get_physical_seg_id(local_seg_id),
-            addr.get_id(),
+            addr.addr_id,
             "!" if wants_ack else ".",
         )
 
     @staticmethod
-    def segment_coupler_scan():
+    def segment_coupler_scan() -> str:
         """Generate a scan-command for LCN segment-couplers.
 
         Used to detect the local segment (where the physical bus connection is
@@ -232,7 +280,7 @@ class PckGenerator:
         return "SK"
 
     @staticmethod
-    def request_serial():
+    def request_serial() -> str:
         """Generate a firmware/serial-number request.
 
         :return: The PCK command (without address header) as text
@@ -241,8 +289,8 @@ class PckGenerator:
         return "SN"
 
     @staticmethod
-    def request_name(block_id):
-        """ Generate a name request.
+    def request_name(block_id: int) -> str:
+        """Generate a name request.
 
         :return The PCK command (without address header) as text
         :rtype:    str
@@ -252,8 +300,8 @@ class PckGenerator:
         return "NMN{:d}".format(block_id + 1)
 
     @staticmethod
-    def request_comment(block_id):
-        """ Generate a comment request.
+    def request_comment(block_id: int) -> str:
+        """Generate a comment request.
 
         :return The PCK command (without address header) as text
         :rtype:    str
@@ -263,8 +311,8 @@ class PckGenerator:
         return "NMK{:d}".format(block_id + 1)
 
     @staticmethod
-    def request_oem_text(block_id):
-        """ Generate an oem text request.
+    def request_oem_text(block_id: int) -> str:
+        """Generate an oem text request.
 
         :return The PCK command (without address header) as text
         :rtype:    str
@@ -274,7 +322,25 @@ class PckGenerator:
         return "NMO{:d}".format(block_id + 1)
 
     @staticmethod
-    def request_output_status(output_id):
+    def request_group_membership_static() -> str:
+        """Generate a group membership request for static membership (EEPROM).
+
+        :return The PCK command (without address header) as text
+        :rtype:    str
+        """
+        return "GP"
+
+    @staticmethod
+    def request_group_membership_dynamic() -> str:
+        """Generate a group membership request for dynamic membership.
+
+        :return The PCK command (without address header) as text
+        :rtype:    str
+        """
+        return "GD"
+
+    @staticmethod
+    def request_output_status(output_id: int) -> str:
         """Generate an output-port status request.
 
         :param    int    output_id:    Output id 0..3
@@ -286,11 +352,11 @@ class PckGenerator:
         return "SMA{:d}".format(output_id + 1)
 
     @staticmethod
-    def dim_ouput(output_id, percent, ramp):
+    def dim_output(output_id: int, percent: float, ramp: int) -> str:
         """Generate a dim command for a single output-port.
 
-        :param    int    output_it:    Output id 0..3
-        :param    int    percent:      Brightness in percent 0..100
+        :param    int    output_id:    Output id 0..3
+        :param    float  percent:      Brightness in percent 0..100
         :param    int    ramp:         Ramp value
         :return:    The PCK command (without address header) as text
         :rtype:    str
@@ -309,42 +375,42 @@ class PckGenerator:
         return pck
 
     @staticmethod
-    def dim_all_outputs(percent, ramp, is1805=False):
+    def dim_all_outputs(percent: float, ramp: int, software_serial: int) -> str:
         """Generate a dim command for all output-ports.
 
-        :param    int    percent:    Brightness in percent 0..100
-        :param    int    ramp:       Ramp value
-        :param    bool   is1805:     True if the target module's firmware is
-                                     180501 or newer, otherwise False
+        :param    float  percent:           Brightness in percent 0..100
+        :param    int    ramp:              Ramp value
+        :param    int    software_serial:   The expected firmware version of all
+                                            receiving modules.
         :return:    The PCK command (without address header) as text
         :rtype:    str
         """
-        percent_round = round(percent * 2)
-        if is1805:
+        percent_round = int(round(percent * 2))
+        if software_serial >= 0x180501:
             # Supported since LCN-PCHK 2.61
             pck = "OY{0:03d}{0:03d}{0:03d}{0:03d}{1:03d}".format(percent_round, ramp)
         elif percent_round == 0:  # All off
-            pck = "AA{:03d}".format(ramp)
+            pck = f"AA{ramp:03d}"
         elif percent_round == 200:  # All on
-            pck = "AE{:03d}".format(ramp)
+            pck = f"AE{ramp:03d}"
         else:
             # This is our worst-case: No high-res, no ramp
-            pck = "AH{:03d}".format(percent_round / 2)
+            pck = "AH{:03d}".format(int(percent_round / 2))
         return pck
 
     @staticmethod
-    def rel_output(output_id, percent):
+    def rel_output(output_id: int, percent: float) -> str:
         """Generate a command to change the value of an output-port.
 
         :param    int    output_id:    Output id 0..3
-        :param    int    percent:      Relative percentage -100..100
+        :param    float  percent:      Relative percentage -100..100
         :return:    The PCK command (without address header) as text
         :rtype:    str
         """
         if (output_id < 0) or (output_id > 3):
             raise ValueError("Invalid output_id.")
 
-        percent_round = round(percent * 2)
+        percent_round = int(round(percent * 2))
         if percent_round % 2 == 0:
             # Use the percent command (supported by all LCN-PCHK versions)
             pck = "A{:d}{:s}{:03d}".format(
@@ -356,10 +422,10 @@ class PckGenerator:
             pck = "O{:d}{:s}{:03d}".format(
                 output_id + 1, "AD" if percent >= 0 else "SB", abs(percent_round)
             )
-            return pck
+        return pck
 
     @staticmethod
-    def toggle_output(output_id, ramp):
+    def toggle_output(output_id: int, ramp: int) -> str:
         """Generate a command that toggles a single output-port.
 
         Toggle mode: (on->off, off->on).
@@ -374,7 +440,7 @@ class PckGenerator:
         return "A{:d}TA{:03d}".format(output_id + 1, ramp)
 
     @staticmethod
-    def toggle_all_outputs(ramp):
+    def toggle_all_outputs(ramp: int) -> str:
         """Generate a command that toggles all output-ports.
 
         Toggle mode: (on->off, off->on).
@@ -383,10 +449,10 @@ class PckGenerator:
         :return:    The PCK command (without address header) as text
         :rtype:    str
         """
-        return "AU{:03d}".format(ramp)
+        return f"AU{ramp:03d}"
 
     @staticmethod
-    def request_relays_status():
+    def request_relays_status() -> str:
         """Generate a relays-status request.
 
         :return: The PCK command (without address header) as text
@@ -395,7 +461,7 @@ class PckGenerator:
         return "SMR"
 
     @staticmethod
-    def control_relays(states):
+    def control_relays(states: List[lcn_defs.RelayStateModifier]) -> str:
         """Generate a command to control relays.
 
         :param     RelayStateModifier    states:    The 8 modifiers for the
@@ -412,7 +478,34 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def control_motors_relays(states):
+    def control_relays_timer(
+        time_msec: int, states: List[lcn_defs.RelayStateModifier]
+    ) -> str:
+        """Generate a command to control relays.
+
+        :param     int                   time_msec: Duration of timer in
+                                                    milliseconds
+        :param     RelayStateModifier    states:    The 8 modifiers for the
+                                                    relay states as a list
+                                                    (only ON and OFF allowed)
+        :return:   The PCK command (without address header) as text
+        :rtype:    str
+        """
+        if len(states) != 8:
+            raise ValueError("Invalid states length.")
+        value = lcn_defs.time_to_native_value(time_msec)
+        ret = f"R8T{value:03d}"
+        for state in states:
+            assert state in (
+                lcn_defs.RelayStateModifier.ON,
+                lcn_defs.RelayStateModifier.OFF,
+            )
+            ret += state.value
+
+        return ret
+
+    @staticmethod
+    def control_motors_relays(states: List[lcn_defs.MotorStateModifier]) -> str:
         """Generate a command to control motors via relays.
 
         :param    MotorStateModifier    states:    The 4 modifiers for the
@@ -449,7 +542,10 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def control_motors_outputs(state, reverse_time=None):
+    def control_motors_outputs(
+        state: lcn_defs.MotorStateModifier,
+        reverse_time: Optional[lcn_defs.MotorReverseTime] = None,
+    ) -> str:
         """Generate a command to control a motor via output ports 1+2.
 
         :param    MotorStateModifier    state:     The modifier for the
@@ -491,7 +587,7 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def request_bin_sensors_status():
+    def request_bin_sensors_status() -> str:
         """Generate a binary-sensors status request.
 
         :return:    The PCK command (without address header) as text
@@ -500,7 +596,7 @@ class PckGenerator:
         return "SMB"
 
     @staticmethod
-    def var_abs(var, value):
+    def var_abs(var: lcn_defs.Var, value: int) -> str:
         """Generate a command that sets a variable absolute.
 
         :param    Var    var:    The target variable to set
@@ -523,7 +619,7 @@ class PckGenerator:
         raise ValueError("Wrong variable type.")
 
     @staticmethod
-    def update_status_var(var, value):
+    def update_status_var(var: lcn_defs.Var, value: int) -> str:
         """Generate a command that send variable status updates.
 
         PCHK provides this variables by itself on selected segments
@@ -540,25 +636,25 @@ class PckGenerator:
             x2cmd = var_id | 0x40
             byte1 = (value >> 8) & 0xFF
             byte2 = value & 0xFF
-            return "X2{:03d}{:03d}{:03d}".format(x2cmd, byte1, byte2)
+            return f"X2{x2cmd:03d}{byte1:03d}{byte2:03d}"
 
         # Setting variables and thresholds absolute not implemented in LCN
         # firmware yet
         raise ValueError("Wrong variable type.")
 
     @staticmethod
-    def var_reset(var, is2013=True):
+    def var_reset(var: lcn_defs.Var, software_serial: int) -> str:
         """Generate a command that resets a variable to 0.
 
-        :param    Var    var:    The target variable to set 0
-        :param    bool   is2013: True if the target module's firmware version
-                                 is 170101 or newer, otherwise False
+        :param    Var    var:               The target variable to set 0
+        :param    int    software_serial:   The expected firmware version of all
+                                            receiving modules.
         :return:    The PCK command (without address header) as text
         :rtype:    str
         """
         var_id = lcn_defs.Var.to_var_id(var)
         if var_id != -1:
-            if is2013:
+            if software_serial >= 0x170206:
                 pck = "Z-{:03d}{:04d}".format(var_id + 1, 4090)
             else:
                 if var_id == 0:
@@ -579,16 +675,20 @@ class PckGenerator:
         raise ValueError("Wrong variable type.")
 
     @staticmethod
-    def var_rel(var, rel_var_ref, value, is2013=True):
+    def var_rel(
+        var: lcn_defs.Var,
+        rel_var_ref: lcn_defs.RelVarRef,
+        value: int,
+        software_serial: int,
+    ) -> str:
         """Generate a command to change the value of a variable.
 
-        :param    Var          var:    The target variable to change
-        :param    RelVarRef    rel_var_ref:   The reference-point
-        :param    int          value:  The native LCN value to add/subtract
-                                       (can be negative)
-        :param    bool         is2013: True if the target module's firmware
-                                       version is 170101 or newer, otherwise
-                                       False
+        :param    Var       var:                The target variable to change
+        :param    RelVarRef rel_var_ref:        The reference-point
+        :param    int       value:              The native LCN value to add/subtract
+                                                (can be negative)
+        :param    int       software_serial:    The expected firmware version of all
+                                                receiving modules.
         :return:    The PCK command (without address header) as text
         :rtype:    str
         """
@@ -619,7 +719,7 @@ class PckGenerator:
         thrs_register_id = lcn_defs.Var.to_thrs_register_id(var)
         thrs_id = lcn_defs.Var.to_thrs_id(var)
         if (thrs_register_id != -1) and (thrs_id != -1):
-            if is2013:
+            if software_serial >= 0x170206:
                 # New command for registers 1-4 (since 170206, LCN-PCHK 2.8)
                 pck = "SS{:s}{:04d}{:s}R{:d}{:d}".format(
                     "R" if rel_var_ref == lcn_defs.RelVarRef.CURRENT else "E",
@@ -630,7 +730,7 @@ class PckGenerator:
                 )
             elif thrs_register_id == 0:
                 # Old command for register 1 (before 170206)
-                pck = "SS{:s}{:4d}{:s}{:s}{:s}{:s}{:s}{:s}".format(
+                pck = "SS{:s}{:04d}{:s}{:s}{:s}{:s}{:s}{:s}".format(
                     "R" if rel_var_ref == lcn_defs.RelVarRef.CURRENT else "E",
                     abs(value),
                     "A" if value >= 0 else "S",
@@ -645,15 +745,15 @@ class PckGenerator:
         raise ValueError("Wrong variable type.")
 
     @staticmethod
-    def request_var_status(var, sw_age=0x170206):
+    def request_var_status(var: lcn_defs.Var, software_serial: int) -> str:
         """Generate a variable value request.
 
-        :param    Var    var:    The variable to request
-        :param    int    swAge:  The target module's firmware version
+        :param    Var    var:               The variable to request
+        :param    int    software_serial:   The target module's firmware version
         :return:  The PCK command (without address header) as text
         :rtype:   str
         """
-        if sw_age >= 0x170206:
+        if software_serial >= 0x170206:
             var_id = lcn_defs.Var.to_var_id(var)
             if var_id != -1:
                 return "MWT{:03d}".format(var_id + 1)
@@ -694,7 +794,7 @@ class PckGenerator:
         raise ValueError("Wrong variable type.")
 
     @staticmethod
-    def request_leds_and_logic_ops():
+    def request_leds_and_logic_ops() -> str:
         """Generate a request for LED and logic-operations states.
 
         :return: The PCK command (without address header) as text
@@ -703,7 +803,7 @@ class PckGenerator:
         return "SMT"
 
     @staticmethod
-    def control_led(led_id, state):
+    def control_led(led_id: int, state: lcn_defs.LedStatus) -> str:
         """Generate a command to the set the state of a single LED.
 
         :param    int          led_id:   Led id 0..11
@@ -713,13 +813,13 @@ class PckGenerator:
         """
         if (led_id < 0) or (led_id > 11):
             raise ValueError("Bad led_id.")
-        return "LA{:03d}{:2}".format(led_id + 1, state.value)
+        return "LA{:03d}{:s}".format(led_id + 1, state.value)
 
     @staticmethod
-    def send_keys(cmds, keys):
+    def send_keys(cmds: List[lcn_defs.SendKeyCommand], keys: List[bool]) -> str:
         """Generate a command to send LCN keys.
 
-        :param    SendKeyCOmmand    cmds:    The 4 concrete commands to send
+        :param    SendKeyCommand    cmds:    The 4 concrete commands to send
                                              for the tables (A-D) as list
         :param    list(bool)        keys:    The tables' 8 key-states (True
                                              means "send") as list
@@ -744,7 +844,9 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def send_keys_hit_deferred(table_id, time, time_unit, keys):
+    def send_keys_hit_deferred(
+        table_id: int, time: int, time_unit: lcn_defs.TimeUnit, keys: List[bool]
+    ) -> str:
         """Generate a command to send LCN keys deferred / delayed.
 
         :param     int         table_id:    Table id 0(A)..3(D)
@@ -760,10 +862,10 @@ class PckGenerator:
         ret = "TV"
         try:
             ret += PckGenerator.TABLE_NAMES[table_id]
-        except IndexError:
-            raise ValueError("Wrong table_id.")
+        except IndexError as exc:
+            raise ValueError("Wrong table_id.") from exc
 
-        ret += "{:03d}".format(time)
+        ret += f"{time:03d}"
         if time_unit == lcn_defs.TimeUnit.SECONDS:
             if (time < 1) or (time > 60):
                 raise ValueError("Wrong time.")
@@ -789,7 +891,7 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def request_key_lock_status():
+    def request_key_lock_status() -> str:
         """Generate a request for key-lock states.
 
         Always requests table A-D. Supported since LCN-PCHK 2.8.
@@ -800,7 +902,7 @@ class PckGenerator:
         return "STX"
 
     @staticmethod
-    def lock_keys(table_id, states):
+    def lock_keys(table_id: int, states: List[lcn_defs.KeyLockStateModifier]) -> str:
         """Generate a command to lock keys.
 
         :param     int           table_id:  Table id 0(A)..3(D)
@@ -812,8 +914,8 @@ class PckGenerator:
             raise ValueError("Bad table_id or states.")
         try:
             ret = "TX{:s}".format(PckGenerator.TABLE_NAMES[table_id])
-        except IndexError:
-            raise ValueError("Wrong table_id.")
+        except IndexError as exc:
+            raise ValueError("Wrong table_id.") from exc
 
         for state in states:
             ret += state.value
@@ -821,7 +923,9 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def lock_keys_tab_a_temporary(time, time_unit, keys):
+    def lock_keys_tab_a_temporary(
+        time: int, time_unit: lcn_defs.TimeUnit, keys: List[bool]
+    ) -> str:
         """Generate a command to lock keys for table A temporary.
 
         There is no hardware-support for locking tables B-D.
@@ -834,8 +938,8 @@ class PckGenerator:
         :rtype:    str
         """
         if len(keys) != 8:
-            raise ValueError("Wrong keys lenght.")
-        ret = "TXZA{:03d}".format(time)
+            raise ValueError("Wrong keys length.")
+        ret = f"TXZA{time:03d}"
 
         if time_unit == lcn_defs.TimeUnit.SECONDS:
             if (time < 1) or (time > 60):
@@ -862,7 +966,7 @@ class PckGenerator:
         return ret
 
     @staticmethod
-    def dyn_text_part(row_id, part_id, part):
+    def dyn_text_part(row_id: int, part_id: int, part: bytes) -> bytes:
         """Generate the command header / start for sending dynamic texts.
 
         Used by LCN-GTxD periphery (supports 4 text rows).
@@ -872,10 +976,10 @@ class PckGenerator:
 
         :param    int    row_id:    Row id 0..3
         :param    int    part_id:   Part id 0..4
-        :param    str    part:      Text part (up to 12 bytes), encoded as
+        :param    bytes  part:      Text part (up to 12 bytes), encoded as
                                     lcn_defs.LCN_ENCODING
-        :return:  The PCK command (without address header) as text
-        :rtype:   str
+        :return:  The PCK command (without address header) as encoded bytes
+        :rtype:   bytes
         """
         if (
             (row_id < 0)
@@ -885,12 +989,10 @@ class PckGenerator:
             or (len(part) > 12)
         ):
             raise ValueError("Wrong row_id, part_id or part length.")
-        return "GTDT{:d}{:d}{:s}".format(
-            row_id + 1, part_id + 1, part.decode(lcn_defs.LCN_ENCODING)
-        )
+        return "GTDT{:d}{:d}".format(row_id + 1, part_id + 1).encode("utf-8") + part
 
     @staticmethod
-    def lock_regulator(reg_id, state):
+    def lock_regulator(reg_id: int, state: bool) -> str:
         """Generate a command to lock a regulator.
 
         :param    int    reg_id:    Regulator id 0..1
@@ -903,8 +1005,8 @@ class PckGenerator:
         return "RE{:s}X{:s}".format("A" if reg_id == 0 else "B", "S" if state else "A")
 
     @staticmethod
-    def change_scene_register(register_id):
-        """Change the scene register id.
+    def change_scene_register(register_id: int) -> str:
+        """Change the active scene register.
 
         :param    int    register_id:    Register id 0..9
         :return:  The PCK command (without address header) as text
@@ -912,10 +1014,38 @@ class PckGenerator:
         """
         if (register_id < 0) or (register_id > 9):
             raise ValueError("Wrong register_id.")
-        return "SZW{:03d}".format(register_id)
+        return f"SZW{register_id:03d}"
 
     @staticmethod
-    def activate_scene_output(scene_id, output_ports=(), ramp=None):
+    def store_scene_outputs_direct(
+        register_id: int, scene_id: int, percents: Sequence[float], ramps: Sequence[int]
+    ) -> str:
+        """Store the given output values and ramps in the given scene.
+
+        :param    int           register_id: Register id 0..9
+        :param    int           scene_id:    Scene id 0..9
+        :param    list(float)   percents:    Output values in percent as list
+        :param    list(int)     ramp:        Ramp values as list
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        if (scene_id < 0) or (scene_id > 9):
+            raise ValueError("Wrong scene_id.")
+        if len(percents) not in (2, 4):
+            raise ValueError("Need 2 or 4 output percent values.")
+        if len(ramps) != len(percents):
+            raise ValueError("Need as many ramp values as output percent values.")
+        cmd = f"SZD{register_id:03d}{scene_id:03d}"
+        for i, percent in enumerate(percents):
+            cmd += f"{int(percent * 2):03d}{ramps[i]:03d}"
+        return cmd
+
+    @staticmethod
+    def activate_scene_output(
+        scene_id: int,
+        output_ports: Sequence[lcn_defs.OutputPort] = (),
+        ramp: Optional[int] = None,
+    ) -> str:
         """Activate the stored output states for the given scene.
 
         Please note: The output ports 3 and 4 can only be activated
@@ -929,6 +1059,40 @@ class PckGenerator:
         :return:  The PCK command (without address header) as text
         :rtype:   str
         """
+        return PckGenerator._activate_or_store_scene_output(
+            scene_id, output_ports, ramp, store=False
+        )
+
+    @staticmethod
+    def store_scene_output(
+        scene_id: int,
+        output_ports: Sequence[lcn_defs.OutputPort] = (),
+        ramp: Optional[int] = None,
+    ) -> str:
+        """Store the current output states in the given scene.
+
+        Please note: The output ports 3 and 4 can only be stored
+        simultaneously. If one of them is given, the other one is stored,
+        too.
+
+        :param    int                scene_id:       Scene id 0..9
+        :param    list(OutputPort)   output_ports:   Output ports to store
+                                                     as list
+        :param    int                ramp:           Ramp value
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        return PckGenerator._activate_or_store_scene_output(
+            scene_id, output_ports, ramp, store=True
+        )
+
+    @staticmethod
+    def _activate_or_store_scene_output(
+        scene_id: int,
+        output_ports: Sequence[lcn_defs.OutputPort] = (),
+        ramp: Optional[int] = None,
+        store: bool = False,
+    ) -> str:
         if (scene_id < 0) or (scene_id > 9):
             raise ValueError("Wrong scene_id.")
         if not output_ports:
@@ -943,14 +1107,20 @@ class PckGenerator:
             or lcn_defs.OutputPort.OUTPUT4 in output_ports
         ):
             output_mask += 4
-        if ramp is None:
-            pck = "SZA{:1d}{:03d}".format(output_mask, scene_id)
+        if store:
+            action = "S"
         else:
-            pck = "SZA{:1d}{:03d}{:03d}".format(output_mask, scene_id, ramp)
+            action = "A"
+        if ramp is None:
+            pck = f"SZ{action:s}{output_mask:1d}{scene_id:03d}"
+        else:
+            pck = f"SZ{action:s}{output_mask:1d}{scene_id:03d}{ramp:03d}"
         return pck
 
     @staticmethod
-    def activate_scene_relay(scene_id, relay_ports=()):
+    def activate_scene_relay(
+        scene_id: int, relay_ports: Sequence[lcn_defs.RelayPort] = ()
+    ) -> str:
         """Activate the stored relay states for the given scene.
 
         :param    int                scene_id:       Scene id 0..9
@@ -959,6 +1129,32 @@ class PckGenerator:
         :return:  The PCK command (without address header) as text
         :rtype:   str
         """
+        return PckGenerator._activate_or_store_scene_relay(
+            scene_id, relay_ports, store=False
+        )
+
+    @staticmethod
+    def store_scene_relay(
+        scene_id: int, relay_ports: Sequence[lcn_defs.RelayPort] = ()
+    ) -> str:
+        """Store the current relay states in the given scene.
+
+        :param    int                scene_id:       Scene id 0..9
+        :param    list(RelayPort)    relay_ports:    Relay ports to store
+                                                     as list
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        return PckGenerator._activate_or_store_scene_relay(
+            scene_id, relay_ports, store=True
+        )
+
+    @staticmethod
+    def _activate_or_store_scene_relay(
+        scene_id: int,
+        relay_ports: Sequence[lcn_defs.RelayPort] = (),
+        store: bool = False,
+    ) -> str:
         if (scene_id < 0) or (scene_id > 9):
             raise ValueError("Wrong scene_id.")
         if not relay_ports:
@@ -966,5 +1162,48 @@ class PckGenerator:
         relays_mask = ["0"] * 8
         for port in relay_ports:
             relays_mask[port.value] = "1"
-        return "SZA0{:03d}{:s}".format(scene_id, "".join(relays_mask))
+        if store:
+            action = "S"
+        else:
+            action = "A"
+        return "SZ{:s}0{:03d}{:s}".format(action, scene_id, "".join(relays_mask))
 
+    @staticmethod
+    def request_status_scene(register_id: int, scene_id: int) -> str:
+        """Request the stored output and ramp values for the given scene.
+
+        :param    int    register_id:    Register id 0..9
+        :param    int    register_id:    Scene id 0..9
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        if (register_id < 0) or (register_id > 9):
+            raise ValueError("Wrong register_id.")
+        if (scene_id < 0) or (scene_id > 9):
+            raise ValueError("Wrong scene_id.")
+        return f"SZR{register_id:03d}{scene_id:03d}"
+
+    @staticmethod
+    def beep(sound: lcn_defs.BeepSound, count: int) -> str:
+        """Make count number of beep sounds.
+
+        :param    BeepSound sound:  Beep sound style
+        :param    int       count:  Number of beeps (1..15)
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        if (count < 1) or (count > 15):
+            raise ValueError("Wrong number of beeps.")
+        return f"PI{sound.value:s}{count:03d}"
+
+    @staticmethod
+    def empty() -> str:
+        """Generate an empty command (LEER) that does nothing.
+
+        Combine with request for acknowledgement to discover and
+        ping modules and to discover and verify group memberships.
+
+        :return:  The PCK command (without address header) as text
+        :rtype:   str
+        """
+        return "LEER"
